import 'dart:convert';
import 'dart:io';

import '../../../../../config_builder/models/config.dart';
import '../../../../models/file_report.dart';
import '../../../metrics/metric_utils.dart';
import '../../../metrics/metrics_list/cyclomatic_complexity/cyclomatic_complexity_metric.dart';
import '../../../metrics/metrics_list/maximum_nesting_level/maximum_nesting_level_metric.dart';
import '../../../metrics/metrics_list/number_of_methods_metric.dart';
import '../../models/reporter.dart';
import '../../utility_selector.dart';
import 'code_climate_issue.dart';

// Code Climate Engine Specification https://github.com/codeclimate/platform/blob/master/spec/analyzers/SPEC.md

/// Creates reports in Code Climate format widely understood by various CI and analysis tools
class CodeClimateReporter implements Reporter {
  final IOSink _output;

  final Config reportConfig;

  /// If true will report in GitLab Code Quality format
  final bool gitlabCompatible;

  CodeClimateReporter(
    this._output, {
    required this.reportConfig,
    this.gitlabCompatible = false,
  });

  @override
  Future<void> report(Iterable<FileReport> records) async {
    if (records.isEmpty) {
      return;
    }

    final codeClimateRecords = records.map(_toIssues).expand((r) => r);

    if (gitlabCompatible) {
      _output.writeln(json.encode(codeClimateRecords.toList()));
    } else {
      for (final record in codeClimateRecords) {
        _output.writeln('${json.encode(record)}\x00');
      }
    }
  }

  Iterable<CodeClimateIssue> _toIssues(FileReport record) {
    final result = <CodeClimateIssue>[
      ...record.classes.keys.expand((key) {
        final component = record.classes[key]!;
        final report = UtilitySelector.componentReport(component);

        return [
          if (isReportLevel(report.methodsCount.level))
            CodeClimateIssue.numberOfMethods(
              component.location.start.line,
              component.location.end.line,
              report.methodsCount.value,
              record.relativePath,
              key,
              readThreshold<int>(
                reportConfig.metrics,
                NumberOfMethodsMetric.metricId,
                10,
              ),
            ),
        ];
      }),
    ];

    return result
      ..addAll(_functionMetrics(record))
      ..addAll(record.issues.map((issue) =>
          CodeClimateIssue.fromCodeIssue(issue, record.relativePath)))
      ..addAll(record.antiPatternCases.map((antiPattern) =>
          CodeClimateIssue.fromDesignIssue(antiPattern, record.relativePath)));
  }

  Iterable<CodeClimateIssue> _functionMetrics(FileReport record) {
    final issues = <CodeClimateIssue>[];

    for (final key in record.functions.keys) {
      final func = record.functions[key]!;
      final report = UtilitySelector.functionReport(func);

      if (isReportLevel(report.cyclomaticComplexity.level)) {
        issues.add(CodeClimateIssue.cyclomaticComplexity(
          func,
          report.cyclomaticComplexity.value,
          record.relativePath,
          key,
          readThreshold<int>(
            reportConfig.metrics,
            CyclomaticComplexityMetric.metricId,
            20,
          ),
        ));
      }

      if (isReportLevel(report.maintainabilityIndex.level)) {
        issues.add(CodeClimateIssue.maintainabilityIndex(
          func,
          report.maintainabilityIndex.value.toInt(),
          record.relativePath,
          key,
        ));
      }

      if (isReportLevel(report.maximumNestingLevel.level)) {
        issues.add(CodeClimateIssue.maximumNestingLevel(
          func,
          report.maximumNestingLevel.value,
          record.relativePath,
          key,
          readThreshold<int>(
            reportConfig.metrics,
            MaximumNestingLevelMetric.metricId,
            5,
          ),
        ));
      }
    }

    return issues;
  }
}
